@@ -1,5 +1,5 @@
 # orm/state.py
-# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -32,7 +32,7 @@ from .. import util
 
 
 @inspection._self_inspects
-class InstanceState(interfaces.InspectionAttr):
+class InstanceState(interfaces.InspectionAttrInfo):
     """tracks state information at the instance level.
 
     The :class:`.InstanceState` is a key object used by the
@@ -47,7 +47,7 @@ class InstanceState(interfaces.InspectionAttr):
     status within a particular :class:`.Session` and details
     about data on individual attributes.  The public API
     in order to acquire a :class:`.InstanceState` object
-    is to use the :func:`.inspect` system::
+    is to use the :func:`_sa.inspect` system::
 
         >>> from sqlalchemy import inspect
         >>> insp = inspect(some_mapped_object)
@@ -62,7 +62,7 @@ class InstanceState(interfaces.InspectionAttr):
     key = None
     runid = None
     load_options = util.EMPTY_SET
-    load_path = ()
+    load_path = PathRegistry.root
     insert_order = None
     _strong_obj = None
     modified = False
@@ -72,6 +72,7 @@ class InstanceState(interfaces.InspectionAttr):
     _orphaned_outside_of_session = False
     is_instance = True
     identity_token = None
+    _last_known_values = ()
 
     callables = ()
     """A namespace where a per-state loader callable can be associated.
@@ -233,6 +234,18 @@ class InstanceState(interfaces.InspectionAttr):
             and self.session_id in sessionlib._sessions
         )
 
+    def _track_last_known_value(self, key):
+        """Track the last known value of a particular key after expiration
+        operations.
+
+        .. versionadded:: 1.3
+
+        """
+
+        if key not in self._last_known_values:
+            self._last_known_values = dict(self._last_known_values)
+            self._last_known_values[key] = NO_VALUE
+
     @property
     @util.dependencies("sqlalchemy.orm.session")
     def session(self, sessionlib):
@@ -260,7 +273,7 @@ class InstanceState(interfaces.InspectionAttr):
         """Return the mapped identity of the mapped object.
         This is the primary key identity as persisted by the ORM
         which can always be passed directly to
-        :meth:`.Query.get`.
+        :meth:`_query.Query.get`.
 
         Returns ``None`` if the object has no primary key identity.
 
@@ -299,7 +312,7 @@ class InstanceState(interfaces.InspectionAttr):
 
     @util.memoized_property
     def mapper(self):
-        """Return the :class:`.Mapper` used for this mapepd object."""
+        """Return the :class:`_orm.Mapper` used for this mapped object."""
         return self.manager.mapper
 
     @property
@@ -335,21 +348,13 @@ class InstanceState(interfaces.InspectionAttr):
             if persistent:
                 if to_transient:
                     if persistent_to_transient is not None:
-                        obj = state.obj()
-                        if obj is not None:
-                            persistent_to_transient(session, obj)
+                        persistent_to_transient(session, state)
                 elif persistent_to_detached is not None:
-                    obj = state.obj()
-                    if obj is not None:
-                        persistent_to_detached(session, obj)
+                    persistent_to_detached(session, state)
             elif deleted and deleted_to_detached is not None:
-                obj = state.obj()
-                if obj is not None:
-                    deleted_to_detached(session, obj)
+                deleted_to_detached(session, state)
             elif pending and pending_to_transient is not None:
-                obj = state.obj()
-                if obj is not None:
-                    pending_to_transient(session, obj)
+                pending_to_transient(session, state)
 
             state._strong_obj = None
 
@@ -453,6 +458,7 @@ class InstanceState(interfaces.InspectionAttr):
                 "load_options",
                 "class_",
                 "expired_attributes",
+                "info",
             )
             if k in self.__dict__
         )
@@ -480,6 +486,8 @@ class InstanceState(interfaces.InspectionAttr):
         self.parents = state_dict.get("parents", {})
         self.modified = state_dict.get("modified", False)
         self.expired = state_dict.get("expired", False)
+        if "info" in state_dict:
+            self.info.update(state_dict["info"])
         if "callables" in state_dict:
             self.callables = state_dict["callables"]
 
@@ -588,6 +596,11 @@ class InstanceState(interfaces.InspectionAttr):
             collection = dict_.pop(k)
             collection._sa_adapter.invalidated = True
 
+        if self._last_known_values:
+            self._last_known_values.update(
+                (k, dict_[k]) for k in self._last_known_values if k in dict_
+            )
+
         for key in self.manager._all_key_set.intersection(dict_):
             del dict_[key]
 
@@ -607,10 +620,17 @@ class InstanceState(interfaces.InspectionAttr):
                 self.expired_attributes.add(key)
                 if callables and key in callables:
                     del callables[key]
-            old = dict_.pop(key, None)
-            if impl.collection and old is not None:
+            old = dict_.pop(key, NO_VALUE)
+            if impl.collection and old is not NO_VALUE:
                 impl._invalidate_collection(old)
 
+            if (
+                self._last_known_values
+                and key in self._last_known_values
+                and old is not NO_VALUE
+            ):
+                self._last_known_values[key] = old
+
             self.committed_state.pop(key, None)
             if pending:
                 pending.pop(key, None)
@@ -714,6 +734,9 @@ class InstanceState(interfaces.InspectionAttr):
                         previous = attr.copy(previous)
                 self.committed_state[attr.key] = previous
 
+            if attr.key in self._last_known_values:
+                self._last_known_values[attr.key] = NO_VALUE
+
         # assert self._strong_obj is None or self.modified
 
         if (self.session_id and self._strong_obj is None) or not self.modified:
@@ -849,12 +872,22 @@ class AttributeState(object):
 
     @property
     def history(self):
-        """Return the current pre-flush change history for
+        """Return the current **pre-flush** change history for
         this attribute, via the :class:`.History` interface.
 
         This method will **not** emit loader callables if the value of the
         attribute is unloaded.
 
+        .. note::
+
+            The attribute history system tracks changes on a **per flush
+            basis**. Each time the :class:`.Session` is flushed, the history
+            of each attribute is reset to empty.   The :class:`.Session` by
+            default autoflushes each time a :class:`_query.Query` is invoked.
+            For
+            options on how to control this, see :ref:`session_flushing`.
+
+
         .. seealso::
 
             :meth:`.AttributeState.load_history` - retrieve history
@@ -866,12 +899,21 @@ class AttributeState(object):
         return self.state.get_history(self.key, PASSIVE_NO_INITIALIZE)
 
     def load_history(self):
-        """Return the current pre-flush change history for
+        """Return the current **pre-flush** change history for
         this attribute, via the :class:`.History` interface.
 
         This method **will** emit loader callables if the value of the
         attribute is unloaded.
 
+        .. note::
+
+            The attribute history system tracks changes on a **per flush
+            basis**. Each time the :class:`.Session` is flushed, the history
+            of each attribute is reset to empty.   The :class:`.Session` by
+            default autoflushes each time a :class:`_query.Query` is invoked.
+            For
+            options on how to control this, see :ref:`session_flushing`.
+
         .. seealso::
 
             :attr:`.AttributeState.history`
